unit mainbtest;
// (C) 2009-2020 Marco van de Voort.

// Generic looptiling version image to rotate image.
// see https://stackoverflow.com/questions/848025/rotating-bitmaps-in-code  (non generic one)

// for 8-bit, the SSE version is even faster:
// discussion: https://stackoverflow.com/questions/47478010/sse2-8x8-byte-matrix-transpose-code-twice-as-slow-on-haswell-then-on-ivy-bridge
// final code: http://www.stack.nl/~marcov/rot8x8.txt

// REQUIRES FPC 3.2.0+

{ Unit test for a comparison between
   - 90 degree image rotation hardcoded for 3 byte images using basic coordinate inversion.
   - A generic version for all n-byte types using 64x64 looptiling.

 Speed improvement is image size dependent. Usually the biggest speedup is for images in the
 neighbourhood of the cache size.
}

{$mode delphi}
{$pointermath on}
{$modeswitch advancedrecords on}

interface

uses
  Types, Windows,Classes, SysUtils, Forms, Controls, Graphics, Dialogs, ExtCtrls, lcltype,
  StdCtrls;

Type
{ This is a simple abstraction that I use to decouple optimized (often assembler) image routines
  from the exact classtype. It assumes that an image is a block of memory that can be parameterized
  by the fields of this record, with some classtype specific code to glue.
  Such abstraction is fairly universal and works for e.g. (industrial) camera images, opengl
  textures and bitmaps.
}

  TImageAbstraction = record
                    bpp        : Integer;  // bytes per pixel
                    baseptr    : pbyte;    // address of pixel on line 0. See converrsion routine
                    rowpitch   : Integer;  // scanline[1]-scanline[0]. Can be negative.
                    ImageWidth : Integer;  // Imagewidth. Can be smaller than abs(rowpitch) due to bmp alignment to 4 bytes.
                    ImageHeight: Integer;  // Imageheight
                    function GetImagePointer(X,Y:integer):pbyte;  inline;
                  end;


  { TForm1 }

  TForm1 = class(TForm)
    butRotatAlign: TButton;
    butRotate90: TButton;
    Image1: TImage;
    Memo1: TMemo;
    procedure RotatAlignClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure Rotate90Click(Sender: TObject);
  private
   srcbitmap,
   destbitmap : TBitmap;
  public

  end;

var
  Form1: TForm1;

implementation

{$R *.lfm}

{$inline on}
{$O+}
{$optimization level4}


// the tbitmap to imageabstraction conversion.
procedure loadbitmap(var x : TImageAbstraction;b:tbitmap);
begin
   x.ImageWidth:=b.Width;
   x.ImageHeight:=b.Height;
   x.rowpitch:=b.Scanline[1]-b.ScanLine[0]; // width of a row pixel in bytes + alignment. can be negative.
   x.baseptr:=b.scanline[0]; // assume topdown. First line in mem is first line of img.
   if x.rowpitch<0 then // bottomup ? -> first line of image is last line in mem.
     x.baseptr:=@pbyte(x.baseptr)[(x.imageheight-1)*abs(x.rowpitch)];
   case b.PixelFormat of
    pf8bit  : x.bpp:=1;
    pf15bit,
    pf16bit : x.bpp:=2;
    pf24bit : x.bpp:=3;
    pf32bit : x.bpp:=4;
   end;
end;

{ TForm1 }

procedure TForm1.FormCreate(Sender: TObject);
begin

end;


function TImageAbstraction.GetImagePointer(X, Y: integer): pbyte;
begin
  Result := pbyte(@pbyte(baseptr)[rowpitch*y+x*bpp])
end;

procedure rotatealign<T>({$ifdef fpc}constref{$else}const [ref] {$endif}Source,Target:TImageAbstraction);

const stepsize = 64; // rotate 64x64 blocks at a time.

Type RefT = ^T;

// warning: this method has some fixes that hardcode it to RGB24 to avoid delphi XE bugs.
// (multiplications by 3)
var stepsx,stepsy,restx,resty : Integer;
   RowPitchSource, RowPitchTarget : Integer;
   pSource, pTarget,ps1,ps2 : RefT;
   x,y,i,j: integer;
   rpstep : integer;
begin
  RowPitchSource := source.RowPitch;                 // bytes to jump to next line. Can be negative (includes alignment)
  RowPitchTarget := target.RowPitch;        rpstep:=RowPitchTarget*stepsize;
  stepsx:=source.ImageWidth div stepsize;
  stepsy:=source.ImageHeight div stepsize;
  // check if mod 16=0 here for both dimensions, if so -> SSE2?
  for y :=    0 to stepsy- 1 do
    begin
      psource:=reft(Source.GetImagePointer(0,y*stepsize));          // gets pointer to pixel x,y
      ptarget:=reft(Target.GetImagePointer(target.imagewidth-(y+1)*stepsize,0));
      for x :=  0 to stepsx - 1 do
        begin
          for i := 0 to stepsize - 1 do
            begin
              ps1:=reft(@pbyte(psource)[rowpitchsource*i]);   // ( 0,i)
              ps2:=reft(@ptarget[(stepsize-1-i)]);       //  (maxx-i,0);
              for j := 0 to stepsize - 1 do
               begin
                 ps2[0]:=ps1[j];
                 inc(pbyte(ps2),RowPitchTarget);
               end;
            end;
          inc(psource,stepsize);
          inc(pbyte(ptarget),rpstep);
        end;
    end;
  // 3 more areas to do, with dimensions
  // - stepsy*stepsize * restx        // right most column of restx width
  // - stepsx*stepsize * resty        // bottom row with resty height
  // - restx*resty                    // bottom-right rectangle.
  restx:=source.ImageWidth mod stepsize;   // typically zero because width is typically 1024 or 2048
  resty:=source.Imageheight mod stepsize;
  if restx>0 then
    begin
      // one loop less, since we know this fits in one line of  "blocks"
      psource:=reft(source.GetImagePointer(source.ImageWidth-restx,0));    // gets pointer to pixel x,y
      ptarget:=reft(Target.GetImagePointer(Target.imagewidth-stepsize,Target.imageheight-restx));
      for y := 0 to stepsy - 1 do
        begin
          for i := 0 to stepsize - 1 do
            begin
              ps1:=@pbyte(psource)[rowpitchsource*i];   // ( 0,i)
              ps2:=@ptarget[stepsize-1-i];       //  (maxx-i,0);
              for j := 0 to restx - 1 do
               begin
                 ps2[0]:=ps1[j];
                 inc(pbyte(ps2),RowPitchTarget);
               end;
            end;
         inc(pbyte(psource),stepsize*RowPitchSource);
         dec(ptarget,stepsize);
       end;
    end;
  if resty>0 then
    begin
      // one loop less, since we know this fits in one line of  "blocks"
      psource:=reft(source.GetImagePointer(0,source.ImageHeight-resty));    // gets pointer to pixel x,y
      ptarget:=reft(Target.GetImagePointer(0,0));
      for x := 0 to stepsx - 1 do
        begin
          for i := 0 to resty- 1 do
            begin
              ps1:=@pbyte(psource)[rowpitchsource*i];   // ( 0,i)
              ps2:=@ptarget[resty-1-i];       //  (maxx-i,0);
              for j := 0 to stepsize - 1 do
               begin
                 ps2[0]:=ps1[j];
                 inc(pbyte(ps2),RowPitchTarget);
               end;
            end;
         inc(psource,stepsize);
         inc(pbyte(ptarget),rpstep);
       end;
    end;
 if (resty>0) and (restx>0) then
    begin
      // one loop less, since we know this fits in one line of  "blocks"
      psource:=reft(source.GetImagePointer(source.ImageWidth-restx,source.ImageHeight-resty));    // gets pointer to pixel x,y
      ptarget:=reft(Target.GetImagePointer(0,target.ImageHeight-restx));
      for i := 0 to resty- 1 do
        begin
          ps1:=@pbyte(psource)[rowpitchsource*i];   // ( 0,i)
          ps2:=@ptarget[resty-1-i];       //  (maxx-i,0);
          for j := 0 to restx - 1 do
            begin
              ps2[0]:=ps1[j];
              inc(pbyte(ps2),RowPitchTarget);
            end;
       end;
    end;
end;

procedure Rotate90(src,dest: TBitmap);
type
  TRGBArray = array[0..0] of TRGBTriple;
  pRGBArray = ^TRGBArray;
var
  oldRows, oldColumns: integer;
  rowIn, rowOut: pRGBArray;
  tmpBitmap: TBitmap;
begin
  for oldColumns := 0 to src.Width - 1 do
  begin
    rowOut := dest.ScanLine[oldColumns];

    for oldRows := 0 to src.Height - 1 do
    begin
      rowIn := src.ScanLine[oldRows];
      rowOut[src.Height - oldRows - 1] := rowIn[oldColumns];
    end;
  end;

end;

// while these routines have a non Windows eq, I haven't really tested their precision.
// most testing on *nix was about magnitude only.
procedure starttimer(var tmr:int64);
begin
  {$ifdef mswindows}
  queryperformancecounter(tmr);
  {$else}
   tmr:=GetTickCount64;
  {$endif}
end;
{$ifdef mswindows}
function  getfreq:int64;
begin
  queryperformancefrequency(result);
end;
{$endif}

function getusdiffs(time2,tmr:int64):single; inline;
begin
  {$ifdef mswindows}
  result:= ((time2-tmr)*1000000) / getfreq;
  {$else}
  result := (time2-tmr)*1000; // hack. Use getclock_monotonic in time?
  {$endif}
end;

function  endtimeruss(tmr : int64):string;
var microsec: Single;
    time2:int64;
begin
  starttimer(time2);
  microsec:=getusdiffs(time2,tmr);
  result:=FloatToStrF(microsec,ffFixed,5,2);
end;

procedure TForm1.RotatAlignClick(Sender: TObject);
var srca,desta : TImageAbstraction;
    tmr : int64;
begin
 if not assigned(srcbitmap) then
    srcbitmap:=image1.Picture.Bitmap;
 if not assigned(destbitmap) then
   begin
     destbitmap :=TBitmap.Create;
     destbitmap.PixelFormat:=srcbitmap.PixelFormat;
     destbitmap.SetSize(srcbitmap.Height,srcbitmap.Width);
   end;
 loadbitmap(srca,srcbitmap);
 loadbitmap(desta,destbitmap);
 starttimer(tmr);
 case destbitmap.PixelFormat of
  pf24bit : rotatealign<TRGBTriple>(srca,desta);
  pf32bit : rotatealign<TRGBQuad>(srca,desta);
 end;
  memo1.lines.add('rotatealign:'+endtimeruss(tmr));
 destbitmap.SaveToFile('test.bmp');
end;

procedure TForm1.Rotate90Click(Sender: TObject);
var tmr : int64;
begin
  if not assigned(srcbitmap) then
     srcbitmap:=image1.Picture.Bitmap;
  if not assigned(destbitmap) then
    begin
      destbitmap :=TBitmap.Create;
      destbitmap.PixelFormat:=srcbitmap.PixelFormat;
      destbitmap.SetSize(srcbitmap.Height,srcbitmap.Width);
    end;
  starttimer(tmr);
  case destbitmap.PixelFormat of
   pf24bit : Rotate90(srcbitmap,destbitmap);
  end;
  memo1.lines.add('rotate90:'+endtimeruss(tmr));
  destbitmap.SaveToFile('test2.bmp');
 end;

end.

